#include "xbasic_types.h"
#include "xio.h"

#include "xintc_l.h"
#include "xuartlite_l.h"

#include "devices.h"

void setup_interrupts()
{

  // reset the interrupt controller

  // this disables the INTC irq output, so the INTC and CPU are "disconnected"
  XIntc_mMasterDisable(XPAR_INTC_SINGLE_BASEADDR);

  // disable all interrupt sources
  XIntc_mEnableIntr(XPAR_INTC_SINGLE_BASEADDR,0); // disable all

  // ack all sources, to be sure none of them is pending due to circuit wakeup
  XIntc_mAckIntr(XPAR_INTC_SINGLE_BASEADDR, 0xffffffff); // ack all sources



  // initialize the drivers and interrupt handlers
  // here we are also supposed to reset the uart etc. TO DO

  // AUDIO AK4565 DRIVER SETUP

  XIntc_InterruptVectorTable[XPAR_INTC_AK4565_INTERRUPT_INTR].Handler=audio_handler;
  audio_dac_bcnt_w=audio_dac_bcnt_r=0;

  // UART DRIVER & INTERRUPT SETUP
  XIntc_InterruptVectorTable[XPAR_INTC_MYUART_INTERRUPT_INTR].Handler=uart_handler;
  uart_rx_bcnt_w=uart_rx_bcnt_r=0; // init pointers to the start of buffer



  // enable the interrupts

  // enable the CPU interrupts
  microblaze_enable_interrupts();

  // enable the INTC irq output 
  XIntc_mMasterEnable(XPAR_INTC_SINGLE_BASEADDR); // master enable

  // enable the INTC to receive interrupts from the AK4565 and UART
  XIntc_mEnableIntr(XPAR_INTC_SINGLE_BASEADDR, 
		    XPAR_AK4565_INTERRUPT_MASK | XPAR_MYUART_INTERRUPT_MASK);
  
  // enable the UART to generate interrupts
  // the audio driver always generate interrupts
  XUartLite_mEnableIntr(XPAR_MYUART_BASEADDR);
 
}


//int sndbeeplen, sndchirplen;

static Xuint16 tri_wave[1024];
void make_triangle_wave()
{
  int i;
  for(i=0;i<512;i++) tri_wave[i]=tri_wave[1023-i]=(i-256)<<7;
}

int sndbeeplen, sndchirplen;

void tetris_sound_init()
{

  int i;
  make_triangle_wave();
  for(i=0;i<1024;i++) audio_dac_buf[i]=0;

}


static int phase = 0;
void tetris_sound()
{

  int i;
  Xuint32 l,r;
  
  unsigned int freq;
  
  while(audio_dac_bcnt_w - audio_dac_bcnt_r > 512);
  
  for(i=0;i<512;i++){

    freq=0;

    if(sndbeeplen > 0){
      sndbeeplen--;
      freq=5000;
    }

    if(sndchirplen > 0){
      sndchirplen--;
      freq = 6000 - sndchirplen;
    };

    l=r=tri_wave[(phase>>8) & 0x3FF];
    phase+=freq;

    audio_dac_buf[audio_dac_bcnt_w++ & 0x3FF] =  r;
  }

}
