////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2004 Xilinx, Inc.  All rights reserved.
//
// Xilinx, Inc.
// XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
// COURTESY TO YOU.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
// ONE POSSIBLE   IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
// STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
// IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
// FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
// XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
// THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
// ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
// FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.
//
// File   : arp.c
// Date   : 2002, March 20.
// Author : Sathya Thammanur
// Company: Xilinx
// Group  : Emerging Software Technologies
//
// Summary:
// ARP layer specific functions
//
// $Id: arp.c,v 1.2 2004/07/28 18:26:51 sathya Exp $
//
////////////////////////////////////////////////////////////////////////////////

#include "net/xilsock.h"
#include "net/eth.h"

/*
 * xilnet_arp: handler for arp packets
 */

void xilnet_arp_reply(unsigned char *buf, unsigned int len);


unsigned int xilnet_arp(unsigned char *buf, unsigned int len) {

  struct xilnet_arp_pkt* ar = (struct xilnet_arp_pkt*) (buf+ETH_HDR_LEN);

  // Analyse if its a arp request and send a reply if needed
  if (ar->hdr.hard_type == ARP_HARD_TYPE_ETH && ar->hdr.prot_type == ETH_PROTO_IP &&
      ar->hdr.hard_size == ARP_HARD_SIZE && ar->hdr.prot_size == ARP_PROTO_IP_SIZE &&
      (ar->target_ip[0] == mb_ip_addr[0] && ar->target_ip[1] == mb_ip_addr[1] &&
       ar->target_ip[2] == mb_ip_addr[2] && ar->target_ip[3] == mb_ip_addr[3]) ) {
    if (ar->hdr.op == ARP_REQ) {
      // update hw addr table
      xilnet_eth_update_hw_tbl(buf, ETH_PROTO_ARP);
      xilnet_arp_reply(buf+ETH_HDR_LEN, len-ETH_HDR_LEN);
    }
  }

   print("xilnet_arp() about to return :");
   putnum(len);
   print("\r\n");
  return len;
}


/*
 * xilnet_arp_reply:
 */

void xilnet_arp_reply(unsigned char *buf, unsigned int len) {

  int i;
  struct xilnet_arp_pkt *pkt;
  struct xilnet_arp_pkt *arp_req = (struct xilnet_arp_pkt*) buf;
  struct xilnet_arp_pkt *arp_reply;
#ifdef _DEBUG_
//    print("arp_reply(): \n");
#endif
  pkt = (struct xilnet_arp_pkt*)(sendbuf+ETH_HDR_LEN);

 //  print("xilnet_arp_reply() :");
 //  putnum(len);
 //  print("\n\r");

  pkt->hdr.hard_type = ARP_HARD_TYPE_ETH;

  /* Modify ETH_PROTO_IP to  ETH_PROTO_ARP if _IP doesn't work */
  /* Sambuddho */
  pkt->hdr.prot_type = ETH_PROTO_IP;
  pkt->hdr.hard_size = ARP_HARD_SIZE;
  pkt->hdr.prot_size = ARP_PROTO_IP_SIZE;
  pkt->hdr.op = ARP_REPLY;

  for (i = 0; i < ETH_ADDR_LEN; i++) {
    pkt->sender_hw[i] = mb_hw_addr[i];
    pkt->target_hw[i] = arp_req->sender_hw[i];
  }

  /* Sambuddho : I will take this out */
//  print(" ARP reply being sent to : ");
//  putnum(arp_req->sender_ip[0]);
//  putnum(arp_req->sender_ip[1]);
//  putnum(arp_req->sender_ip[2]);
//  putnum(arp_req->sender_ip[3]);

   print("\r\n");

  for (i = 0; i < ARP_PROTO_IP_SIZE ; i++) {
  //  print("Writing sender IP:");
    pkt->sender_ip[i] = mb_ip_addr[i];
  //  putnum(pkt->sender_ip[i]);
  //  print("\r\n");
  //  print("Writing target IP:");
    pkt->target_ip[i] = arp_req->sender_ip[i];
  //  putnum(pkt->target_ip[i]);
  //  print("\r\n");
  }

//  memset(pkt->pad, 0, ARP_PAD_SIZE);

  for(i=0;i<ARP_PAD_SIZE;i++)
     pkt->pad[i] = 0;


  /* initialize ethernet header */
/*
  for(i=0;i<ETH_MIN_FRAME_LEN;i++)
  {
  	putnum(sendbuf[i]);
	print (" ");
  }
  */

  xilnet_eth_send_frame(sendbuf, ETH_MIN_FRAME_LEN, arp_req->sender_ip, pkt->target_hw, ETH_PROTO_ARP);

}

